/*
 * File name: cma_mod.c
 *
 * Description: CMA utility module to allow device user access and CMA
 *
 * Copyright (C) 2013 Texas Instruments, Incorporated
 * Author: Aurelien Jacquiot <a-jacquiot@ti.com>
 *
 * Some code borrowed from Linux virtio_udma driver
 * Copyright (C) 2012, 2013 Texas Instruments Incorporated
 * Author: Cyril Chemparathy <cyril@ti.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */
#include "linux/cma_mod.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/dma-mapping.h>
#include <linux/version.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/io.h>
#include <linux/hrtimer.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/of_address.h>
#include <linux/platform_device.h>

#include <asm/cacheflush.h>

#ifndef CONFIG_CMA /* for old kernels */
#define dma_mmap_coherent(dev, vma, cpu_addr, dma_addr, size)  \
   remap_pfn_range(vma, (vma)->vm_start, (vma)->vm_pgoff,   \
                   size, (vma)->vm_page_prot)
#else
#include <linux/dma-contiguous.h>
#endif

#ifndef VM_RESERVED
#define VM_RESERVED  (VM_DONTEXPAND | VM_DONTDUMP)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
static inline void* PDE_DATA(const struct inode *inode)
{
   return PDE(inode)->data;
}
#endif


static atomic_t cma_dev_ref_count = ATOMIC_INIT(0);

/* DDR cached/dmacoherent */
struct cma_range {
   struct list_head list;
   dma_addr_t dma_addr;
   phys_addr_t phys_addr;
   size_t size;
   atomic_t offset;
};

static LIST_HEAD(msmc_ranges);
static LIST_HEAD(ddr_ranges);

struct mpax {
   u32  mpaxl;
   u32  mpaxh;
};

struct msms_regs {
   u32  pid;
   u32  _res_04;
   u32  smcerrar;
   u32  smcerrxr;
   u32  smedcc;
   u32  smcea;
   u32  smsecc;
   u32  smpfar;
   u32  smpfxr;
   u32  smpfr;
   u32  smpfcr;
   u32  _res_2c;
   u32  sbndc[8];
   u32  sbndm;
   u32  sbnde;
   u32  _res_58;
   u32  cfglck;
   u32  cfgulck;
   u32  cfglckstat;
   u32  sms_mpax_lck;
   u32  sms_mpax_ulck;
   u32  sms_mpax_lckstat;
   u32  ses_mpax_lck;
   u32  ses_mpax_ulck;
   u32  ses_mpax_lckstat;
   u32  smestat;
   u32  smirstat;
   u32  smirc;
   u32  smiestat;
   u32  smiec;
   u32  _res_94_c0[12];
   u32  smncerrar;
   u32  smncerrxr;
   u32  smncea;
   u32  _res_d0_1fc[76];
   struct mpax sms[16][8];
   struct mpax ses[16][8];
};

static inline u32 mpax_dma_addr(const struct mpax __iomem *m)
{
   return __raw_readl(&m->mpaxh)&0xfffff000;
}

static inline u64 mpax_size(const struct mpax __iomem *m)
{
   return (u64)2<<(__raw_readl(&m->mpaxh)&0x1f);
}

static inline u64 mpax_phys_addr(const struct mpax __iomem *m)
{
   return ((u64)__raw_readl(&m->mpaxl)&0xffffff00)<<4;
}

static inline bool mpax_enabled(const struct mpax __iomem *m)
{
   return (__raw_readl(&m->mpaxh)&0x1f)?true:false;
}

static inline bool mpax_coherent(const struct mpax __iomem *m)
{
   return (__raw_readl(&m->mpaxh)&0x80)?false:true;
}

#if 0
static void mpax_add(struct mpax __iomem *m, phys_addr_t phys, u32 dma, u64 size)
{
   __raw_writel((dma&0xffffff000)|(ilog2(size)-1), &m->mpaxh);
   __raw_writel(((phys>>4)&0xffffff00)|0x3f, &m->mpaxl);
}
#endif

#define MPU0_BASE_ADDRESS     0x02360000
#define MPU1_BASE_ADDRESS     0x02368000
#define MPU2_BASE_ADDRESS     0x02370000
#define MPU3_BASE_ADDRESS     0x02378000
#define MPU4_BASE_ADDRESS     0x02380000
#define MPU5_BASE_ADDRESS     0x02388000
#define MPU6_BASE_ADDRESS     0x02388400
#define MPU7_BASE_ADDRESS     0x02388800

#define MPU0_RANGE_NUM        16 /* Main CFG SCR/TeraNEt  */
#define MPU1_RANGE_NUM        16 /* QMSS data port */
#define MPU2_RANGE_NUM        16 /* QMSS cfg port 1 */
#define MPU5_RANGE_NUM        16 /* QMSS cfg port 2 */
#define MPU7_RANGE_NUM        16 /* DDR3 */

#define MPU_PROG_MPSAR(n)     (0x200 + ((n) << 4))
#define MPU_PROG_MPEAR(n)     (0x204 + ((n) << 4))
#define MPU_PROG_MPPA(n)      (0x208 + ((n) << 4))

/* copies the supervisor permissions to user permissions.  This allows
 * userspace access while still preventing execute of register space and
 * write of read-only registers */
static void copy_mpu_supervisor_to_user(unsigned long base, unsigned count)
{
   unsigned int i;
   unsigned long reg_addr;

   reg_addr = base + MPU_PROG_MPPA(0);

   for (i = 0; i < count; ++i) {
      u32 __iomem *reg = ioremap(reg_addr, 4);
      u32 val = __raw_readl(reg);
      val = (val & ~0x7) | ((val>>3)&0x7);
      __raw_writel(val, reg);
      iounmap(reg);
      reg_addr += 0x10;
   }
}

static void mpu_enable_user_mode_access(void)
{
   /* QMSS data (VBUSM) */
   copy_mpu_supervisor_to_user(MPU1_BASE_ADDRESS, MPU1_RANGE_NUM);

   /* QMSS cfg port 1 (VBUSP) */
   copy_mpu_supervisor_to_user(MPU2_BASE_ADDRESS, MPU2_RANGE_NUM);

   /* QMSS cfg port 2 (VBUSP) */
   copy_mpu_supervisor_to_user(MPU5_BASE_ADDRESS, MPU5_RANGE_NUM);

   /* DDR3B */
   copy_mpu_supervisor_to_user(MPU7_BASE_ADDRESS, MPU7_RANGE_NUM);
}

/*********************************************************************************
* FUNCTION: cma_mod_read_proc
*
**********************************************************************************
*
* DESCRIPTION: Forms output for /proc/cma_mod file
*********************************************************************************/
static void* cma_seq_start(struct seq_file *m, loff_t *pos)
{
   loff_t off = *pos-1;
   struct list_head *lh;

   m->private = &msmc_ranges;

   if (!*pos) {
      return SEQ_START_TOKEN;
   }
   
   list_for_each(lh, &msmc_ranges) {
      if (off-- == 0) {
         return lh;
      }
   }
   
   m->private = &ddr_ranges;
   return seq_list_start(&ddr_ranges, off);
}

static void* cma_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
   void *next;

   if (v == SEQ_START_TOKEN) {
      ++*pos;
      return seq_list_start(&msmc_ranges, 0);
   }

   next = seq_list_next(v, m->private, pos);

   if (!next && m->private == &msmc_ranges) {
      m->private = &ddr_ranges;
      return seq_list_start(&ddr_ranges, 0);
   }
   
   return next;
}

static void cma_seq_stop(struct seq_file *m, void *v)
{
}

static int cma_seq_show(struct seq_file *m, void *v)
{
   struct cma_range *r;

   if (v == SEQ_START_TOKEN) {
      seq_printf(m, "DMA      PHYSICAL  SIZE     OFFSET\n");
      return 0;
   }
   
   r = list_entry(v, typeof(*r), list);
   
   seq_printf(m, "%08x %09llx %08x %08x\n",
      r->dma_addr, r->phys_addr, r->size, atomic_read(&r->offset));

   return 0;
}

static const struct seq_operations cma_seq_ops = {
   .start = cma_seq_start,
   .next = cma_seq_next,
   .stop = cma_seq_stop,
   .show = cma_seq_show
};

static int cma_mod_proc_open(struct inode *inode, struct file *file)
{
   //((struct seq_file *)file->private_data)->private = PDE_DATA(inode);
   return seq_open(file, &cma_seq_ops);
}

static const struct file_operations cma_mod_proc_fops = {
   .open              = cma_mod_proc_open,
   .read              = seq_read,
   .llseek            = seq_lseek,
   .release           = seq_release,
};



static int mpax_seq_show(struct seq_file *m, void *v)
{
   struct msms_regs __iomem *msmc;
   unsigned privid, idx;

   msmc = ioremap(0x0bc00000, 0x100000);
   if (!msmc) {
      return -ENOMEM;
   }

   seq_printf(m, "ID ## DMA      PHYSICAL  SIZE\n");
   
   for (privid=0; privid<ARRAY_SIZE(msmc->ses); ++privid) {
      for (idx=0; idx<ARRAY_SIZE(msmc->ses[0]); ++idx) {
         struct mpax __iomem *mpax = &msmc->ses[privid][idx];
         
         if (!mpax_enabled(mpax)) {
            continue;
         }
         
         seq_printf(m, "%2u %2u %08x %09llx %8llx %sCOHERENT\n",
            privid, idx, mpax_dma_addr(mpax), mpax_phys_addr(mpax),
            mpax_size(mpax), mpax_coherent(mpax)?"":"IN");
      }
   }
   
   iounmap(msmc);

   return 0;
}

static int mpax_proc_open(struct inode *inode, struct file *file)
{
   return single_open(file, mpax_seq_show, PDE_DATA(inode));
}

static const struct file_operations mpax_proc_fops = {
   .open              = mpax_proc_open,
   .read              = seq_read,
   .llseek            = seq_lseek,
   .release           = single_release,
};

static bool pfn_in_range(unsigned long start_pfn, unsigned long size, const struct cma_range *r)
{
   if (start_pfn >= PFN_DOWN(r->dma_addr) &&
      start_pfn+PFN_UP(size) <= PFN_UP((phys_addr_t)r->dma_addr+r->size))
   {
      return true;
   }
   
   return false;
}

static int cma_mod_mmap(struct file *file, struct vm_area_struct *vma)
{
   struct device *dev = ((struct miscdevice *)file->private_data)->this_device;
   unsigned long    pfn = 0;
   size_t           size;
   int              ret;
   const struct cma_range *r;

   size = vma->vm_end - vma->vm_start;

   dev_dbg(dev, "vma->vm_start     = %#lx\n", vma->vm_start);
   dev_dbg(dev, "vma->vm_end       = %#lx\n", vma->vm_end);
   dev_dbg(dev, "vma->vm_pgoff     = %#lx\n", vma->vm_pgoff);
   dev_dbg(dev, "vma->vm_pgprot    = %#llx\n", pgprot_val(vma->vm_page_prot));
   dev_dbg(dev, "size              = 0x%x\n", size);
   dev_dbg(dev, "paddr             = %#lx\n", (vma->vm_pgoff << PAGE_SHIFT));

   /* ensure we are within a valid range */
   list_for_each_entry(r, &msmc_ranges, list) {
      if (pfn_in_range(vma->vm_pgoff, size, r)) {
         pfn = vma->vm_pgoff - PFN_DOWN(r->dma_addr) + PFN_DOWN(r->phys_addr);
         goto found;
      }
   }
   
   list_for_each_entry(r, &ddr_ranges, list) {
      if (pfn_in_range(vma->vm_pgoff, size, r)) {
         pfn = vma->vm_pgoff - PFN_DOWN(r->dma_addr) + PFN_DOWN(r->phys_addr);
         goto found;
      }
   }
   
   return -ENXIO;

found:

   if (file->f_flags & O_DSYNC) {
      vma->vm_flags |= VM_IO | VM_RESERVED;
      dev_dbg(dev, "D_SYNC -> (noncached/coherent memory)\n");
      vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
   } else {
      dev_dbg(dev, "no D_SYNC -> (cached/dmacoherent memory)\n");
      /* Keystone modifies pgprot_dmacoherent to be cacheable and outer-shareable */
      vma->vm_page_prot = pgprot_dmacoherent(vma->vm_page_prot);
   }

   ret = remap_pfn_range(vma,
                         vma->vm_start,
                         pfn,
                         size,
                         vma->vm_page_prot);

   if (ret) {
      dev_err(dev, "failed to user map dma memory\n");
      return ret;
   }

   return 0;
}

#ifndef CONFIG_MMU
static unsigned long cma_mod_get_unmapped_area(struct file  *file,
      unsigned long addr,
      unsigned long len,
      unsigned long pgoff,
      unsigned long flags)
{
   return pgoff << PAGE_SHIFT;
}
#endif

static long cma_mod_ioc_init(struct device *dev, struct cma_mod_init __user *argp)
{
   struct cma_range *r;

   dev_alert(dev, "INIT called: clearing mappings with %u user(s)\n",
      atomic_read(&cma_dev_ref_count));

   /* Make sure that the caller is the only process that has this
      device opened */
   if (atomic_read(&cma_dev_ref_count) != 1) {
      dev_err(dev, "cma device still in use\n");
      return -EBUSY;
   }

   /* Initialize global data */
   list_for_each_entry(r, &msmc_ranges, list) {
      atomic_set(&r->offset, 0);
   }
   list_for_each_entry(r, &ddr_ranges, list) {
      atomic_set(&r->offset, 0);
   }

   return 0;
}

/* deprecated! */
/* This can return only the first mapping */
static long cma_mod_ioc_getinfo(struct device *dev, struct cma_mod_getinfo __user *argp)
{
   struct cma_mod_getinfo mem_info;
   struct list_head *head;
   struct cma_range *r;

   dev_dbg(dev, "GETINFO ioctl received.\n");

   if (copy_from_user(&mem_info, argp, sizeof(mem_info)))
      return -EFAULT;

   switch (mem_info.mem_type) {
   case CMA_MOD_MEM_TYPE_MSM:
      head = &msmc_ranges;
      break;

   case CMA_MOD_MEM_TYPE_DDR:
      head = &ddr_ranges;
      break;

   default:
      dev_err(dev, "GETINFO ioctl invalid argument mem_type %#x\n",
              mem_info.mem_type);
      return -EINVAL;
   }  /* switch base mem_type */


   if (list_empty(head)) {
      return -ENOMEM;
   }
   
   r = list_entry(head->next, struct cma_range, list);
   mem_info.addr = r->dma_addr;
   mem_info.size = r->size;

   if (copy_to_user(argp, &mem_info, sizeof(mem_info)))
      return -EFAULT;

   dev_dbg(dev, "GETINFO returning addr %#llx size %#x\n",
           (long long)mem_info.addr, mem_info.size);

   return 0;
}

static unsigned atomic_add_le(atomic_t *v, unsigned inc, unsigned size)
{
   unsigned off = atomic_read(v);

   for(;;) {
      unsigned old, new;
      
      new = off + inc;
      
      if (new > size) {
         /* failed */
         return -1UL;
      }
      
      old = atomic_cmpxchg(v, off, new);
      if (off == old) {
         /* success */
         return old;
      }
      
      off = old;
   }
}

static long cma_mod_ioc_alloc(struct device *dev, struct cma_mod_alloc __user *argp)
{
   struct cma_mod_alloc alloc;
   struct cma_range *r;
   struct list_head *head;

   dev_dbg(dev, "ALLOC ioctl received.\n");

   if (copy_from_user(&alloc, argp, sizeof(alloc)))
      return -EFAULT;

   alloc.size = PAGE_ALIGN(alloc.size);

   switch (alloc.mem_type) {
   case CMA_MOD_MEM_TYPE_MSM:
      head = &msmc_ranges;
      break;

   case CMA_MOD_MEM_TYPE_DDR:
      head = &ddr_ranges;
      break;

   default:
      dev_err(dev, "ALLOC ioctl invalid argument mem_type %#x\n",
              alloc.mem_type);
      return -EINVAL;
   }  /* switch base mem_type */

   list_for_each_entry(r, head, list) {
      unsigned off;
      
      off = atomic_add_le(&r->offset, alloc.size, r->size);
      if (off != -1UL) {
         alloc.addr = r->dma_addr + off;
         break;
      }
   }
   
   if (&r->list == head) {
      /* allocation failed */
      return -ENOMEM;
   }

   if (copy_to_user(argp, &alloc, sizeof(alloc)))
      return -EFAULT;

   dev_dbg(dev, "ALLOC(MSM) returning addr %#llx, size %#x\n", (long long)alloc.addr, alloc.size);

   return 0;
}

static long cma_mod_ioctl(struct file *file, unsigned int cmd, unsigned long args)
{
   void __user *argp = (void __user *) args;
   struct device *dev = ((struct miscdevice *)file->private_data)->this_device;

   switch (cmd) {
   case CMA_MOD_IOC_INIT:
      return cma_mod_ioc_init(dev, argp);
   case CMA_MOD_IOC_GETINFO:
      return cma_mod_ioc_getinfo(dev, argp);
   case CMA_MOD_IOC_ALLOC:
      return cma_mod_ioc_alloc(dev, argp);
   case CMA_MOD_IOC_ENABLEPMU:
      break;

   default:
      dev_warn(dev, "Unknown ioctl %#x received.\n", cmd);
      return -ENOTTY;
   } /* switch base cmd */

   return 0;
}

static int cma_mod_open(struct inode *inode, struct file *file)
{
   struct device *dev = ((struct miscdevice *)file->private_data)->this_device;

   dev_dbg(dev, "open: called.\n");

   atomic_inc(&cma_dev_ref_count);
   return 0;
}

static int cma_mod_release(struct inode *inode, struct file *file)
{
   struct device *dev = ((struct miscdevice *)file->private_data)->this_device;

   dev_dbg(dev, "close: called.\n");

   atomic_dec(&cma_dev_ref_count);
   return 0;
}

static const struct file_operations cma_mod_fops = {
   .owner             = THIS_MODULE,
   .open              = cma_mod_open,
   .release           = cma_mod_release,
   .mmap              = cma_mod_mmap,
#ifndef CONFIG_MMU
   .get_unmapped_area = cma_mod_get_unmapped_area,
#endif
   .unlocked_ioctl    = cma_mod_ioctl,
};

static struct miscdevice cma_dev = {
   .minor = MISC_DYNAMIC_MINOR,
   .name  = CMA_MOD_DEVNAME,
   .fops  = &cma_mod_fops,
};

#if 0
/* the DSPs get the first 2 indices */
#define ARM_MPAX_SEGMENT_START_IDX 2

static int add_mpax_mapping(struct device *dev, phys_addr_t phys, u32 dma, u32 size)
{
   struct msms_regs __iomem *msmc;
   unsigned id, i;
   int err = 0;

   msmc = ioremap(0x0bc00000, 0x100000);
   if (!msmc) {
      dev_err(dev, "failed to ioremap\n");
      return -ENOMEM;
   }

   for (id=0; id<ARRAY_SIZE(msmc->ses); ++id) {
      for (i=ARM_MPAX_SEGMENT_START_IDX; i<ARRAY_SIZE(msmc->ses[id]); ++i) {
         if (mpax_enabled(&msmc->ses[id][i])) {
            if (mpax_dma_addr(&msmc->ses[id][i]) == dma) {
               /* found existing mapping at this address */
               dev_dbg(dev, "privid %u replacing entry %u: %08x -> %09llx +%08llx\n",
                  id, i, mpax_dma_addr(&msmc->ses[id][i]),
                  mpax_phys_addr(&msmc->ses[id][i]), mpax_size(&msmc->ses[id][i]));
                  
               mpax_add(&msmc->ses[id][i], phys, dma, size);
               break;
            }
         }
      }

      if (i != ARRAY_SIZE(msmc->ses[id])) {
         /* found and replaced */
         continue;
      }
      
      /* search for first free entry */
      for (i=ARM_MPAX_SEGMENT_START_IDX; i<ARRAY_SIZE(msmc->ses[id]); ++i) {
         if (!mpax_enabled(&msmc->ses[id][i])) {
            dev_dbg(dev, "privid %u adding entry %u\n", id, i);

            mpax_add(&msmc->ses[id][i], phys, dma, size);
            break;
         }
      }
      
      if (i == ARRAY_SIZE(msmc->ses[id])) {
         dev_err(dev, "no free MPAX SES entries for privid %u\n", id);
         err = -ENOSPC;
         goto out;
      }
   }

   err = 0;

out:   
   iounmap(msmc);

   return err;
}

static int configure_mpax(struct device *dev, struct list_head *head)
{
   int err;
   struct cma_range *r;

   list_for_each_entry(r, head, list) {
      dev_dbg(dev, "translate %08x -> %09llx +%08zx\n", r->dma_addr, (u64)r->phys_addr, r->size);
      if ((err = add_mpax_mapping(dev, r->phys_addr, r->dma_addr, r->size))) {
         dev_err(dev, "unable to add translation %08x -> %09llx +%08zx: %d\n", r->dma_addr, (u64)r->phys_addr, r->size, err);
         return err;
      }
   }
   
   return 0;
}
#endif

static int add_ranges(struct device *dev, struct device_node *np, struct list_head *head)
{
   struct cma_range *r;
   const __be32 *addrp;
   u64 size;
   int i = 0;

   while ((addrp = of_get_address(np, i++, &size, NULL))) {
      r = kmalloc(sizeof(*r), GFP_KERNEL);
      if (!r)
         return -ENOMEM;
      
      /* we can assume this is 32-bit w/o running of_n_addr_cells */
      r->dma_addr = be32_to_cpup(addrp);
      r->phys_addr = of_translate_address(np, addrp);;
      r->size = size;
      atomic_set(&r->offset, 0);
      
      list_add_tail(&r->list, head);

      dev_dbg(dev, "range %08x -> %09llx +%08zx\n", r->dma_addr, (u64)r->phys_addr, r->size);
   }
   
   return 0;
}

static void *cma_virt_addr;
static dma_addr_t cma_dma_addr;

static int add_kernel_cma_range(struct device *dev, size_t size)
{
   struct cma_range *r;

   r = kmalloc(sizeof(*r), GFP_KERNEL);
   if (!r)
      return -ENOMEM;

   cma_virt_addr = dma_alloc_coherent(dev, size, &cma_dma_addr, GFP_KERNEL | GFP_DMA);
   if (!cma_virt_addr) {
      dev_err(dev, "error allocating from CMA for size 0x%zx\n",
              size);
      kfree(r);
      return -ENOMEM;
   }

   r->dma_addr = cma_dma_addr;
   r->phys_addr = virt_to_phys(cma_virt_addr);
   r->size = size;
   atomic_set(&r->offset, 0);
   
   list_add_tail(&r->list, &ddr_ranges);
   
   return 0;
}

static void check_del_kernel_cma_range(struct device *dev, struct cma_range *r)
{
   if (r->dma_addr == cma_dma_addr) {
      dma_free_coherent(dev, r->size, cma_virt_addr, r->dma_addr);
   }
}

/*********************************************************************************
* FUNCTION: cma_mod_init_module
*
* DESCRIPTION:Initialization routine for cma kernel device
*********************************************************************************/
static int cma_mod_probe(struct platform_device *pdev)
{
   int err;
	struct device_node *np = pdev->dev.of_node;
	struct device_node *child;
   struct cma_range *r, *rnext;
   u32 cma_size;

   /* enable user access to qmss h/w */
   mpu_enable_user_mode_access();

   child = of_get_child_by_name(np, "msmc");
   if (child) {
      err = add_ranges(&pdev->dev, child, &msmc_ranges);
      of_node_put(child);
      
      if (err)
         goto out;
   }else{
      dev_info(&pdev->dev, "no 'msmc' node\n");
   }
   
   child = of_get_child_by_name(np, "ddr");
   if (child) {
      err = add_ranges(&pdev->dev, child, &ddr_ranges);
      of_node_put(child);

      if (err)
         goto out_ranges;
   }else{
      dev_info(&pdev->dev, "no 'ddr' node\n");
   }

#if 0
   err = configure_mpax(&pdev->dev, &ddr_ranges);
   if (err)
      goto out_ranges;
#endif

   if (!of_property_read_u32(np, "cma_size", &cma_size)) {
      /* allocate DDR cached memory from CMA */
      /* no need for MPAX mapping above since this is from kernel */
      err = add_kernel_cma_range(&pdev->dev, cma_size);
      if (err)
         goto out_ranges;
   }
   
   if (!proc_create(CMA_MOD_DEVNAME, S_IRUGO, NULL, &cma_mod_proc_fops)) {
      err = -ENOMEM;
      goto out_ranges;
   }

   if (!proc_create("mpax", S_IRUGO, NULL, &mpax_proc_fops)) {
      err = -ENOMEM;
      goto out_cma_proc;
   }

   cma_dev.parent = &pdev->dev;
   err = misc_register(&cma_dev);
   if (err) {
      dev_err(&pdev->dev, "could not register misc device\n");
      goto out_mpax_proc;
   }

   dev_dbg(&pdev->dev, "Allocated minor number: %d\n", cma_dev.minor);
   dev_dbg(&pdev->dev, "module loaded\n");

   return 0;

out_mpax_proc:
   remove_proc_entry("mpax", NULL);

out_cma_proc:
   remove_proc_entry(CMA_MOD_DEVNAME, NULL);

out_ranges:
   list_for_each_entry_safe(r, rnext, &msmc_ranges, list) {
      kfree(r);
   }
   list_for_each_entry_safe(r, rnext, &ddr_ranges, list) {
      check_del_kernel_cma_range(&pdev->dev, r);
      kfree(r);
   }

out:
   return err;
}

/*********************************************************************************
 * FUNCTION: cma_mod_cleanup_module
 *
 * DESCRIPTION:
 *********************************************************************************/
static int cma_mod_remove(struct platform_device *pdev)
{
   struct cma_range *r, *rnext;

   if (atomic_read(&cma_dev_ref_count)) {
      dev_err(&pdev->dev, "cma device still in use\n");
   }

   /* Remove cma device */
   misc_deregister(&cma_dev);

   remove_proc_entry("mpax", NULL);

   /* Remove cma proc entry */
   remove_proc_entry(CMA_MOD_DEVNAME, NULL);

   list_for_each_entry_safe(r, rnext, &msmc_ranges, list) {
      kfree(r);
   }
   list_for_each_entry_safe(r, rnext, &ddr_ranges, list) {
      check_del_kernel_cma_range(&pdev->dev, r);
      kfree(r);
   }

   dev_dbg(&pdev->dev, "module unloaded\n");

   return 0;
}


static const struct of_device_id of_match[] = {
	{ .compatible = "nsn,cma", },
	{},
};
MODULE_DEVICE_TABLE(of, of_match);

static struct platform_driver cma_drv = {
   .driver = {
      .name = CMA_MOD_DEVNAME,
      .owner = THIS_MODULE,
      .of_match_table = of_match,
   },
   .probe = cma_mod_probe,
   .remove = cma_mod_remove,
};
module_platform_driver(cma_drv);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nokia Siemens Networks");
MODULE_AUTHOR("Texas Instruments Incorporated");
MODULE_DESCRIPTION("CMA module");
