/*
 * Copyright (c) 2011-2015 Nokia Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/**
 * @file cpld.c
 * @author Nokia Solutions and Networks
 * @remarks Copyright (C) 2011-2014 Nokia Solutions and Networks
 *
 * @par PURPOSE
 * FZM CPLD module.
 *
 * </pre>
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <mod_utils.h>
#include "linux/fzm_ioctls.h"
#include "linux/cpld-usr.h"
#include <linux/delay.h>
#include <linux/version.h>
#include "fzm_board_type.h"
#include "cpld_internal.h"

#define DRV_NAME "cpld-usr"

static u8 cpld_bid;
static u8 cpld_ver;

/* Register attribute used for showing or storing a register value
 * or for showing or storing a bit in a register */
struct reg_attr {
   struct device_attribute attr;
   unsigned offset;
   unsigned bit;    /* bit(mask) to read or write */
};

/* TODO do we really need to do this for the guarded msid/msstat/smstat/mscntl show and store functinos? Better to check board type and populate as needed? */
#define GUARDED_MIN_CPLD_VER 0xA
static int verify_min_cpld_ver(unsigned offset)
{
   if (cpld_bid >= 0x10)
      return 0; /* Version check below is not applicable to BID 0x10 and higher, always return success */

   if (cpld_ver < GUARDED_MIN_CPLD_VER) {
      printk(KERN_ERR "register offset 0x%x is not supported in CPLD version less than 0x%02x (current CPLD version: 0x%02x)\n", offset, GUARDED_MIN_CPLD_VER, cpld_ver);
      return -EINVAL;
   }

   return 0;
}

/* Generic register show (legacy) */
/* For compatibility only - don't use this for future registers
 * since it inappropriately displays in signed decimal format */
static ssize_t reg_show_compat_dontuse(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg;

   if (fzm_cpldspi_read_locked(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg)))
      return -EIO;

   return snprintf(buf, PAGE_SIZE, "%d\n", reg);
}

/* Generic register show (displays in hex) */
static ssize_t reg_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg;

   if (fzm_cpldspi_read_locked(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg)))
      return -EIO;

   return snprintf(buf, PAGE_SIZE, "0x%02x\n", reg);
}

/* Generic register store (also supports RMW and masked writes) */
static ssize_t reg_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg = 0;
   int err;

   if (!n)
      return n;

   cpld_lock();

   if (*buf == '|' || *buf == '&' || *buf == '^') {
      /* read/modify/write */
      u8 val = 0;
      char *next;

      if (fzm_cpldspi_read(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg))) {
         err = -EIO;
         goto error;
      }

      do {
         val = simple_strtoul(buf + 1, &next, 0);

         if (*buf == '|')
            reg |= val;
         else if (*buf == '&')
            reg &= val;
         else if (*buf == '^')
            reg ^= val;

         buf = next;
      } while (*buf);
   } else if (*buf == 'm') {
      /* masked write */
      u8 val, mask;
      char mask_buf[5];
      char *w;

      w = strchr(buf, 'w');
      if (!w)  {
         err = -EINVAL;
         goto error;
      } else {
         /* value to write */
         val = simple_strtoul(w + 1, NULL, 0);

         /* extract mask */
         memset(mask_buf, 0, sizeof(mask_buf));
         /* only support max of 4 characters for mask */
         if (w - (buf + 1) < sizeof(mask_buf)) {
            strncpy(mask_buf, buf + 1, w - (buf + 1));
            mask = simple_strtoul(mask_buf, NULL, 0);

            if (fzm_cpldspi_read(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg))) {
               err = -EIO;
               goto error;
            }

            /* write the bits */
            reg &= mask;
            reg |= val & ~mask;
         } else {
            err = -EINVAL;
            goto error;
         }
      }
   } else {
      /* write only */
      if (kstrtou8(buf, 0, &reg)) {
         err = -EINVAL;
         goto error;
      }
   }

   if (fzm_cpldspi_write(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg))) {
      err = -EIO;
      goto error;
   }

   cpld_unlock();

   return n;

error:
   cpld_unlock();

   return err;
}


/* Generic register show that verifies CPLD version first */
/* TODO do we really need to do this for msid/smstat/mscntl? Better to check board type and populate as needed? */
static ssize_t guarded_reg_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg;
   int retv;

   retv = verify_min_cpld_ver(ra->offset);
   if (retv)
      return retv;

   if (fzm_cpldspi_read_locked(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg)))
      return -EIO;

   return snprintf(buf, PAGE_SIZE, "%d\n", reg);
}

/* Generic register store that verifies CPLD version first */
/* TODO do we really need to do this for msid/smstat/mscntl? Better to check board type and populate as needed? */
static ssize_t guarded_reg_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg;
   int retv;

   retv = verify_min_cpld_ver(ra->offset);
   if (retv)
      return retv;

   if (!n) {
      return -EINVAL;
   }

   if (kstrtou8(buf, 0, &reg))
      return -EINVAL;

   if (fzm_cpldspi_write_locked(FZM_CPLDSPI_CPLD, ra->offset, &reg, 1))
      return -EIO;

   return n;
}

/* Generic bit show */
static ssize_t bit_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg;

   if (fzm_cpldspi_read_locked(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg)))
      return -EIO;

   if (ra->offset == offsetof(struct cpld_regs, rrr2))
      /* RRR2 logic: 0 is out of reset, 1 is in reset
       * but for the attribute we flip this around
       */
      return snprintf(buf, PAGE_SIZE, "%c\n", (reg & ra->bit) ? '0' : '1');
   else
      return snprintf(buf, PAGE_SIZE, "%c\n", (reg & ra->bit) ? '1' : '0');
}

/* Generic bit store */
static ssize_t bit_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
   struct reg_attr *ra = container_of(attr, struct reg_attr, attr);
   u8 reg;
   u8 on;
   bool flip = false;

   if (ra->offset == offsetof(struct cpld_regs, rrr2))
      /* RRR2 logic: 0 is out of reset, 1 is in reset
       * but for the attribute we flip this around
      */
      flip = true;

   if (!n) {
      return -EINVAL;
   }

   switch (buf[0]) {
   case 'y':
   case 'Y':
   case '1':
      if (flip)
         on = 0;
      else
         on = ra->bit;
      break;
   case 'n':
   case 'N':
   case '0':
      if (flip)
         on = ra->bit;
      else
         on = 0;
      break;
   default:
      return -EINVAL;
   }


   cpld_lock();

   if (fzm_cpldspi_read(FZM_CPLDSPI_CPLD, ra->offset, &reg, sizeof(reg)))
      return -EIO;

   reg = (reg & ~ra->bit) | on;

   if (fzm_cpldspi_write(FZM_CPLDSPI_CPLD, ra->offset, &reg, 1))
      return -EIO;

   cpld_unlock();

   return n;
}

static ssize_t led_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   Cpld_led_t led;
   u8 led1 = 0;
   u8 led2 = 0;
   /* GPIO Input Register, used to get rf_led status if it is controlled by the FPGA */
   u8 gir = 0;
   unsigned shift;
   const char *color[4] = { "", "red", "green", "yellow" };

   if (attr->attr.name[0] == 's')
      led = CPLD_LED_STATUS;
   else if (attr->attr.name[0] == 'b')
      led = CPLD_LED_BH;
   else if (attr->attr.name[0] == 'r')
      led = CPLD_LED_RF;
   else if (attr->attr.name[0] == 'o')
      led = CPLD_LED_OFF;
   else
      return -EIO;

   cpld_lock();
   if (CPLD_LED_RF == led) {
      gir = readb_cpld_gir();
      //printk("gir value: 0x%x.\n", gir);
   }
   if (!(gir & 0x10)) {
      led1 = readb_cpld_led1();
      led2 = readb_cpld_led2();
   }
   cpld_unlock();

   /* if bit FPGA0 of gir is set, the rf_led is controlled by FPGA
   * bit FPGA1 = 1 meaning solid green, 0, meaning blinking green, no other color, no off.
   */
   if (gir & 0x10) {
      return snprintf(buf, PAGE_SIZE, "%s %s\n", (gir & 0x20) ? "solid" : "blinking", "green");
   }

   if (led != CPLD_LED_OFF) {
      shift = led * 2;
      led1 >>= shift;
      led1 &= 0x3;
      led2 >>= shift;
      led2 &= 0x3;

      /* printk("<2>led val 1: %x  2: %x\n", led1, led2); */

      if (!led1)
         return snprintf(buf, PAGE_SIZE, "off\n");
      else
         return snprintf(buf, PAGE_SIZE, "%s %s\n", (led2 & 1) ? "blinking" : "solid", color[led1]);
   } else {
      shift = 7;
      led1 >>= shift;
      led1 &= 0x1;

      if (!led1)
         return snprintf(buf, PAGE_SIZE, "enabled\n");
      else
         return snprintf(buf, PAGE_SIZE, "disabled\n");
   }
}

static ssize_t led_store(struct device *dev, struct device_attribute *attr,
                         const char *buf, size_t n)
{
   Cpld_led_t led;
   u8 led1 = 0;
   u8 led2 = 0;
   u8 reg = 0;
   unsigned shift;

   if (attr->attr.name[0] == 's')
      led = CPLD_LED_STATUS;
   else if (attr->attr.name[0] == 'b')
      led = CPLD_LED_BH;
   else if (attr->attr.name[0] == 'r')
      led = CPLD_LED_RF;
   else if (attr->attr.name[0] == 'o')
      led = CPLD_LED_OFF;
   else
      return -EIO;


   if (led != CPLD_LED_OFF) {
      shift = led * 2;

      if (*buf == 'o')
         led1 = 0;
      else if (*buf == 'r')
         led1 = 0x1;
      else if (*buf == 'g')
         led1 = 0x2;
      else if (*buf == 'y')
         led1 = 0x3;
      else
         return -EINVAL;

      led1 <<= shift;

      if (buf[1] == 'b')
         led2 = 0x1;
      else
         led2 = 0;

      led2 <<= shift;
   } else {
      shift = 7;  /* led off is last bit */

      if (*buf == 'e')  /* if enabled write 0, else write 1 */
         led1 = 0x0;
      else
         led1 = 0x1;

      led1 <<= shift;
   }

   cpld_lock();

   /* printk("<2>led write val 1: %x  2: %x\n", led1, led2); */

   if (led != CPLD_LED_OFF) {
      reg = readb_cpld_led1();
      /* printk("<2> reg val %x\n", reg); */
      reg &= ~(0x3 << shift);
      reg |= led1;

      writeb_cpld_led1(reg);

      reg = readb_cpld_led2();
      /* printk("<2> reg val %x\n", reg); */
      reg &= ~(0x3 << shift);
      reg |= led2;
      writeb_cpld_led2(reg);

   } else {
      reg = readb_cpld_led1();
      /* printk("<2> reg val %x\n", reg); */
      reg &= ~(0x1 << shift);
      reg |= led1;
      /* printk("<2> reg val %x\n", reg); */
      writeb_cpld_led1(reg);
   }

   cpld_unlock();
   return n;
}

static u8 save_next_reset = 0;

/*Below function will trigger the reset based on the save_next_reset value */
static void trigger_cpld_reset(void)
{
   if (save_next_reset) {
      writeb_cpld_rrr1(save_next_reset);
   }
}
static ssize_t next_reset_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   return snprintf(buf, PAGE_SIZE, "%d\n", save_next_reset);
}

extern void register_fzm_cpld_reset(void (*trigger_cpld_reset)(void));

/* set the SPI io_type to poll or interrupt based on the save_next_reset value */
/* register_fzm_cpld_reset() function will register the cpld reset function trigger_cpld_reset()
   with the kernel */

static ssize_t next_reset_store(struct device *dev, struct device_attribute *attr,
                                const char *buf, size_t n)
{
   save_next_reset = simple_strtoul(buf, NULL, 0);
   if (save_next_reset) {
      register_fzm_cpld_reset(trigger_cpld_reset);
   } else {
      register_fzm_cpld_reset(NULL);
   }
   return n;
}

static ssize_t rf_presence_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   u8 reg;
   unsigned idx;
   static const char *val[4] = {"none", "wifi", "rf", "invalid"};

   reg = readb_cpld_rfwfstat();
   /* rf[0-2]_presence */
   idx = attr->attr.name[2] - '0';

   return snprintf(buf, PAGE_SIZE, "%s\n", val[(reg>>(idx*2))&3]);
}

#define _REGRW_ATTR(name) \
   { __ATTR(name, 0644, reg_show, reg_store), offsetof(struct cpld_regs, name ) }

#define _REGR_ATTR(name) \
   { __ATTR(name, 0444, reg_show, NULL), offsetof(struct cpld_regs, name), }

/* Don't use the _COMPAT macros for new registers, they use
 * a show function that displays in signed decimal format */
#define _REGRW_ATTR_COMPAT(name) \
   { __ATTR(name, 0644, reg_show_compat_dontuse, reg_store), offsetof(struct cpld_regs, name ) }

#define _REGR_ATTR_COMPAT(name) \
   { __ATTR(name, 0444, reg_show_compat_dontuse, NULL), offsetof(struct cpld_regs, name), }

#define _GREGRW_ATTR(name) \
   { __ATTR(name, 0644, guarded_reg_show, guarded_reg_store), offsetof(struct cpld_regs, name), }

#define _GREGR_ATTR(name) \
   { __ATTR(name, 0444, guarded_reg_show, NULL), offsetof(struct cpld_regs, name), }

#define _BITRW_ATTR(name, reg, bit) \
   { __ATTR(name, 0644, bit_show, bit_store), offsetof(struct cpld_regs, reg), bit }

#define _BITR_ATTR(name, reg, bit) \
   { __ATTR(name, 0444, bit_show, NULL), offsetof(struct cpld_regs, reg), bit }

/* List of attributes common for all FZM variants
 *
 */
static struct reg_attr reg_base_attrs[] = {
   _REGRW_ATTR_COMPAT(sjmpr),
   //_REGRW_ATTR(ctr),
   { __ATTR(gps_power, 0644, reg_show_compat_dontuse, reg_store), offsetof(struct cpld_regs, ctr) }, /* legacy attribute name, should be deprecated in the future */
   //_REGRW_ATTR(wdc, 0xff),
   { __ATTR(wdc_period, 0644, reg_show_compat_dontuse, reg_store), offsetof(struct cpld_regs, wdc) }, /* legacy attribute name, should be deprecated in the future */
   _REGRW_ATTR_COMPAT(sfpc),
   _REGR_ATTR_COMPAT(nmi),
   _GREGR_ATTR(msid),
   _GREGR_ATTR(msstat),
   _GREGR_ATTR(smstat),
   _GREGRW_ATTR(mscntl),
   _REGRW_ATTR(scr1),
   _REGRW_ATTR(scr2),
   _REGRW_ATTR(scr3),
   _REGRW_ATTR(scr4),
   _REGRW_ATTR(scr5),
   _REGRW_ATTR(scr6),
   _REGRW_ATTR(scr7),
   _REGRW_ATTR(scr8),
   _BITRW_ATTR(bh_phy_reset, rrr2, CPLD_RRR2_BIT_BHPHY),
   _BITRW_ATTR(bh_phy2_reset, rrr2, CPLD_RRR2_BIT_BHPHY2),
   _BITR_ATTR(fpga_reset, rrr2, CPLD_RRR2_BIT_FPGA),
   _BITRW_ATTR(pse_reset, rrr2, CPLD_RRR2_BIT_PSE),
   _BITRW_ATTR(ext_switch_reset, rrr2, CPLD_RRR2_BIT_EXT_SWITCH),
   _BITRW_ATTR(ext_wifi_rrr2_reset, rrr2, CPLD_RRR2_BIT_EXT_WIFI),
   _BITRW_ATTR(ext_wifi_reset, wctr, CPLD_WCTR_BIT_SW_RST_WIFI),
   _BITRW_ATTR(ext_wifi_factory_reset, wctr, CPLD_WCTR_BIT_FCTRY_RST_WIFI),
   { __ATTR(next_reset, 0644, next_reset_show, next_reset_store) }, /* doesn't really belong here, it has its own special show and store functions, but it's common to all variants */
};

static struct device_attribute led_attrs[] = {
   __ATTR(st_led, 0644, led_show, led_store),
   __ATTR(bh_led, 0644, led_show, led_store),
   __ATTR(rf_led, 0644, led_show, led_store),
   __ATTR(off_led, 0644, led_show, led_store),
};

/* TODO move the other wifi related attributes here so that they get
 * created only when wifi is supported. For now, keep the legacy behavior.
 */
static struct reg_attr reg_wifi_attrs[] = {
   _BITRW_ATTR(wifi_power, wctr, CPLD_WCTR_BIT_WIFI_PWR),
};

/* List of attributes valid on indoor (FZP) only */
static struct reg_attr reg_fzp_attrs[] = {
   _BITR_ATTR(jcstat, jcstat, CPLD_JCSTAT_BITV),
};

/* List of attributes valid on indoor (FZP) + WiFi only */
static struct reg_attr reg_fzp_wifi_attrs[] = {
   _BITRW_ATTR(bh_phy3_reset, rrr2, CPLD_RRR2_BIT_BHPHY3),
};

/* TODO create reg_3c_attrs and move the 3c related ones here so that
 * they get craeted only when on a 3c board. For now, keep the legacy
 * behavior with the guarded show and store functions.
 */

/* List of attributes valid on argon 4x4 only */
static struct reg_attr reg_4x4_attrs[] = {
   _REGRW_ATTR(rtcr),
   _BITRW_ATTR(bh_phy_reset_sticky, pstkyren1, CPLD_PSTKYREN1_BIT_BHPHY),
   _BITRW_ATTR(bh_phy2_reset_sticky, pstkyren1, CPLD_PSTKYREN1_BIT_BHPHY2),
   _BITRW_ATTR(wifi_phy_reset, rrr2, CPLD_RRR2_BIT_BHPHY3),
   _BITRW_ATTR(wifi_phy_reset_sticky, pstkyren1, CPLD_PSTKYREN1_BIT_BHPHY3),
   _BITRW_ATTR(fpga_reset_sticky, pstkyren1, CPLD_PSTKYREN1_BIT_FPGA),
   _BITRW_ATTR(pse_reset_sticky, pstkyren1, CPLD_PSTKYREN1_BIT_PSE),
   _BITRW_ATTR(ext_switch_reset_sticky, pstkyren1, CPLD_PSTKYREN1_BIT_EXT_SWITCH),
   _BITRW_ATTR(sfp1_txdis_sticky, pstkyren2, CPLD_PSTKYREN2_BIT_TXDIS1),
   _BITRW_ATTR(sfp2_txdis_sticky, pstkyren2, CPLD_PSTKYREN2_BIT_TXDIS2),
   _REGRW_ATTR(cscr1),
   _REGRW_ATTR(cscr2),
   _REGRW_ATTR(cscr3),
   _REGRW_ATTR(cscr4),
};

/* Valid for boards with RF cards (dual-band indoor & outdoor) only */
static struct reg_attr reg_rfcard_attrs[] = {
   { __ATTR(rf0_presence, 0444, rf_presence_show, NULL) },
   { __ATTR(rf1_presence, 0444, rf_presence_show, NULL) },
   { __ATTR(rf2_presence, 0444, rf_presence_show, NULL) },
   _BITRW_ATTR(rf0_eeprom_writeprotect, rfci2cwp, CPLD_RFCI2CWR_BIT_RF0_EEPROM),
   _BITRW_ATTR(rf1_eeprom_writeprotect, rfci2cwp, CPLD_RFCI2CWR_BIT_RF1_EEPROM),
   _BITRW_ATTR(rf2_eeprom_writeprotect, rfci2cwp, CPLD_RFCI2CWR_BIT_RF2_EEPROM),
};

static struct attribute **cpld_attrs;

static struct attribute_group cpld_attr_group;

/*****************/

/****************************************************************************
 *
 * U T I L I T Y     F U N C T I O N S
 *
 ***************************************************************************/
/**
 * @internal
 * @brief Put a device into reset.
 */
static int cpld_reset_device(Cpld_device_t dev)
{
   int ret = 0;
   u8 bit = 0;

   switch (dev) {
   case CPLD_RESET_FPGA:
      /* don't want anyone resetting these devices */
      ret = -EPERM;
      break;
   case CPLD_RESET_BH_PHY:
      bit = CPLD_RRR2_BIT_BHPHY;
      break;
   case CPLD_RESET_BH_PHY2:
      bit = CPLD_RRR2_BIT_BHPHY2;
      break;
   case CPLD_RESET_BH_PHY3:
      bit = CPLD_RRR2_BIT_BHPHY3;
      break;
   case CPLD_RESET_PSE:
      bit = CPLD_RRR2_BIT_PSE;
      break;
      /* no default so the compiler can assert that we handle all enum cases */
   }

   if (bit) {
      cpld_lock();
      writeb_cpld_rrr2(readb_cpld_rrr2() & ~bit);
      cpld_unlock();
   } else if (!ret) {
      ret = -EINVAL;
   }

   return ret;
}

/**
 * @internal
 * @brief Release device from reset.
 */
static int cpld_release_reset_device(Cpld_device_t dev)
{
   int ret = 0;
   u8 bit = 0;

   switch (dev) {
   case CPLD_RESET_FPGA:
      bit = CPLD_RRR2_BIT_FPGA;
      break;
   case CPLD_RESET_BH_PHY:
      bit = CPLD_RRR2_BIT_BHPHY;
      break;
   case CPLD_RESET_BH_PHY2:
      bit = CPLD_RRR2_BIT_BHPHY2;
      break;
   case CPLD_RESET_BH_PHY3:
      bit = CPLD_RRR2_BIT_BHPHY3;
      break;
   case CPLD_RESET_PSE:
      bit = CPLD_RRR2_BIT_PSE;
      break;

      /* no default so the compiler can assert that we handle all enum cases */
   }

   if (bit) {
      cpld_lock();
      writeb_cpld_rrr2(readb_cpld_rrr2() | bit);
      cpld_unlock();
   } else if (!ret) {
      ret = -EINVAL;
   }

   return ret;
}


static int cpld_set_led(struct cpld_led __user *usr)
{
   struct cpld_led led;
   u8 led1 = 0;
   u8 led2 = 0;
   unsigned shift;

   if (copy_from_user(&led, usr, sizeof(led))) {
      MOD_ERROR("copy_from_user failed\n");
      return -EFAULT;
   }

   if (led.led > CPLD_LED_RF) {
      return -EINVAL;
   }

   shift = led.led * 2;

   cpld_lock();
   led1 = readb_cpld_led1();
   led2 = readb_cpld_led2();

   led1 &= ~(0x3 << shift);
   led1 |= (led.red | (led.green << 1)) << shift;

   led2 &= ~(0x3 << shift);
   led2 |= led.blink << shift;

   writeb_cpld_led1(led1);
   writeb_cpld_led2(led2);
   cpld_unlock();

   return 0;
}


static int cpld_get_led(struct cpld_led __user *usr)
{
   struct cpld_led led;
   u8 led1 = 0;
   u8 led2 = 0;
   unsigned shift;

   if (copy_from_user(&led, usr, sizeof(led))) {
      MOD_ERROR("copy_from_user failed\n");
      return -EFAULT;
   }

   if (led.led > CPLD_LED_RF) {
      return -EINVAL;
   }

   cpld_lock();
   led1 = readb_cpld_led1();
   led2 = readb_cpld_led2();
   cpld_unlock();

   shift = led.led * 2;
   led1 >>= shift;
   led2 >>= shift;

   led.red = !!(led1 & 1);
   led.green = !!(led1 & 2);
   led.blink = !!(led2 & 1);

   if (copy_to_user(usr, &led, sizeof(led))) {
      MOD_ERROR("copy_to_user failed\n");
      return -EFAULT;
   }

   return 0;
}

static int cpld_set_gpio(void)
{
   uint8_t reg = 0x40;

   if (cpld_bid >= 0x10) {
      /* This function is only applicable to BID < 0x10 according to HW team */
      return 0;
   }

   if (cpld_ver <= 0x8) {
      cpld_lock();
      writeb_cpld_gor(reg);
      writeb_cpld_gdr(reg);
      cpld_unlock();
   }

   return 0;
}


/****************************************************************************
 *
 * F I L E     O P     I M P L E M E N T A T I O N S
 *
 ***************************************************************************/


/**
 * @internal
 * Called for IOCTLs on the device.
 */
static long cpldusr_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
   void __user *usr_ptr = (void __user *)arg;
   uint8_t val = 0;
   long ret = 0;

   switch (cmd) {
   case CPLD_RESET_DEVICE:
      ret = cpld_reset_device((Cpld_device_t)arg);
      break;
   case CPLD_RELEASE_RESET_DEVICE:
      ret = cpld_release_reset_device((Cpld_device_t)arg);
      break;
   case CPLD_RD_RESET_STATUS:
      cpld_lock();
      val = readb_cpld_rrr2();
      cpld_unlock();

      if (copy_to_user(usr_ptr, &val, sizeof(val))) {
         ret = -EFAULT;
      }
      break;
   case CPLD_SET_LED:
      ret = cpld_set_led(usr_ptr);
      break;
   case CPLD_GET_LED:
      ret = cpld_get_led(usr_ptr);
      break;
   case CPLD_SET_SCR5:
      val = (uint8_t __user)arg;
      cpld_lock();
      writeb_cpld_scr5(val);
      cpld_unlock();
      break;
   case CPLD_GET_SCR5:
      cpld_lock();
      val = readb_cpld_scr5();
      cpld_unlock();

      if (copy_to_user(usr_ptr, &val, sizeof(val))) {
         ret = -EFAULT;
      }

      break;
   case CPLD_TOGGLE_SCR5: {
      struct cpld_toggle_scr5 tog;

      if (copy_from_user(&tog, usr_ptr, sizeof(tog))) {
         ret = -EFAULT;
      }

      cpld_lock();
      val = readb_cpld_scr5();
      val |= tog.turn_on;
      val &= ~tog.turn_off;
      writeb_cpld_scr5(val);
      cpld_unlock();

      break;
   }

   case CPLD_GET_OFFSET: {
      struct cpld_ioctl_rw rw;

      if (copy_from_user(&rw, usr_ptr, sizeof(rw))) {
         return -EFAULT;
      }

      if (rw.offset % 4 != 0) {
         MOD_ERROR("CPLD_GET_OFFSET: invalid cpld offset\n");
         return -EINVAL;
      }

      cpld_lock();
      rw.value = readb_cpld_offset(rw.offset);
      cpld_unlock();

      if (copy_to_user(usr_ptr, &rw, sizeof(rw))) {
         ret = -EFAULT;
      }
      break;
   }

   case CPLD_SET_OFFSET: {
      struct cpld_ioctl_rw rw;

      if (copy_from_user(&rw, usr_ptr, sizeof(rw))) {
         return -EFAULT;
      }

      if (rw.offset % 4 != 0) {
         MOD_ERROR("CPLD_SET_OFFSET: invalid cpld offset\n");
         return -EINVAL;
      }

      cpld_lock();
      writeb_cpld_offset(rw.offset, rw.value);
      cpld_unlock();
      break;
   }

   default:
      /* pass down to SFP ioctl handler */
      ret = cpld_sfp_ioctl(file, cmd, arg);
   }

   return ret;
}

/***** Char dev operations *****/
static ssize_t cpld_usr_read(struct file *file, char __user *usrbuf, size_t len, loff_t *poff)
{
   size_t rem;
   size_t off = *poff;
   int reg;

   /* ensure size and alignment */
   if (off > CPLD_SIZE) {
      return -EFAULT;
   }

   len = min(len, CPLD_SIZE - off);

   for (rem = len; rem; rem--) {
      reg = readb_cpld_offset(off);
      if (reg < 0) {
         return reg;
      }

      if (put_user(reg, usrbuf)) {
         return -EFAULT;
      }

      usrbuf++;
      off++;
   }

   *poff = off;
   return len;
}

static ssize_t cpld_usr_write(struct file *file, const char __user *usrbuf, size_t len, loff_t *poff)
{
   size_t rem;
   size_t off = *poff;
   unsigned int reg;
   int err;

   /* ensure size and alignment */
   if (off > CPLD_SIZE) {
      return -EFAULT;
   }

   len = min(len, CPLD_SIZE - off);

   for (rem = len; rem; rem--) {
      if (get_user(reg, usrbuf)) {
         return -EFAULT;
      }

      err = writeb_cpld_offset(off, reg);
      if (err) {
         return err;
      }

      usrbuf++;
      off++;
   }

   *poff = off;
   return len;
}

/**
 * Called when acquiring/releasing a lock with flock().
 */
static int cpld_usr_flock(struct file *fp, int cmd, struct file_lock *fl)
{
   int err;

   switch (fl->fl_type) {
   case F_RDLCK:
   case F_WRLCK:
      if (IS_SETLKW(cmd)) {
         if (cpld_lock_interruptible()) {
            return -ERESTARTSYS;
         }
      } else if (!cpld_trylock()) {
         return -EWOULDBLOCK;
      }

      err = flock_lock_file_wait(fp, fl);
      if (err) {
         cpld_unlock();
      }
      break;
   case F_UNLCK:
      fl->fl_flags |= FL_EXISTS;
      err = flock_lock_file_wait(fp, fl);
      if (!err) {
         cpld_unlock();
      }
      break;
   default:
      return -EIO;
   }

   return err;
}

static struct file_operations fops = {
   .owner = THIS_MODULE,
   .read = cpld_usr_read,
   .write = cpld_usr_write,
   .flock = cpld_usr_flock,
   .llseek = default_llseek,
   .unlocked_ioctl = cpldusr_ioctl,
};


/**
 * @internal
 * @brief Proc-fs read handler. Dumps CPLD info.
 *
 * @param seq seq file to read
 * @param v unused
 *
 * @return 0 on success, else errno.
 */
static int cpldusr_proc_read(struct seq_file *seq, void *v)
{
   int i;

   static struct {
      uint16_t reg_offset;
      char *reg_name;
   } debug_regs[] = {
#define REG(x)    { offsetof(struct cpld_regs, x), #x },
#define REGW(x)   { 0xffff, #x },
#define REGR(x)   REG(x)
#define REGG(x)   REGW(x)
#define RESV(n)
#  include <fzm_cpld_regs.def>
   };

   /* printk("<2> Read!!!!\n"); */

   for (i = 0; i < ARRAY_SIZE(debug_regs); ++i) {
      uint8_t reg = 0;

      switch (debug_regs[i].reg_offset) {
      case 0xffff:
         seq_printf(seq, "%-10s = <writeonly>\n", debug_regs[i].reg_name);
         break;

      default:
         fzm_cpldspi_read(FZM_CPLDSPI_CPLD,
                          debug_regs[i].reg_offset, &reg,
                          sizeof(reg));
         seq_printf(seq, "%-10s = 0x%02x\n", debug_regs[i].reg_name, reg);
         break;
      }
   }

   cpld_sfp_proc_read(seq, v);

   return 0;
}

/**
 * @internal
 * @brief Proc-fs open for read handler.
 */
static int cpldusr_proc_open(struct inode *inode, struct file *file)
{
   return single_open(file, cpldusr_proc_read, PDE_DATA(inode));
}

static struct file_operations proc_ops = {
   .owner = THIS_MODULE,
   .open = cpldusr_proc_open,
   .read = seq_read,
   .llseek = seq_lseek,
   .release = single_release,
};


static struct cdev cpldusr_cdev;
static struct proc_dir_entry *cpldusr_proc;
static struct device *dev;

static struct class cpld_class = {
      .name = DRV_NAME,
      .owner = THIS_MODULE,
   };

/**
 * @internal
 * @brief Initializes the module.
 *
 * @return 0 on success.
 */
int cpld_usr_probe(struct device *pdev)
{
   int err;
   dev_t devt;
   int total_num_attrs, i;
   struct attribute **cpld_attrs_p;

#ifdef CPLD_DEBUG
   printk(KERN_EMERG "Inside cpldusr_probe\n");
#endif

   cpld_bid = readb_cpld_bid();
   cpld_ver = readb_cpld_ver();

   err = alloc_chrdev_region(&devt, 0, 1, DRV_NAME);
   if (err) {
      MOD_ERROR("alloc_chrdev_region failed\n");
      return err;
   }

   cdev_init(&cpldusr_cdev, &fops);
   err = cdev_add(&cpldusr_cdev, devt, 1);
   if (err)
      goto out_chrdev_region;

   err = class_register(&cpld_class);
   if (err)
      goto out_cdev;

   dev = device_create(&cpld_class, pdev,
                       devt, NULL, DRV_NAME);
   if (IS_ERR(dev)) {
      err = PTR_ERR(dev);
      goto out_class;
   }

   /* Tally up total number of attributes and allocate an array of attribute
    * pointers */
   total_num_attrs = ARRAY_SIZE(reg_base_attrs) + ARRAY_SIZE(led_attrs);

   /* Note: on dual band boards, WIFI_PWR_ENB (WCTR) doesn't actually
    * control WiFi power, but we use it as a scratchpad register to keep
    * track of WiFi power state. We keep the attribute name the same to avoid
    * board type checks elsewhere. */
   if (is_wifi_supported())
      total_num_attrs += ARRAY_SIZE(reg_wifi_attrs);

   if (is_fzp_board())
      total_num_attrs += ARRAY_SIZE(reg_fzp_attrs);

   if (is_fzp_board() && is_wifi_supported())
      total_num_attrs += ARRAY_SIZE(reg_fzp_wifi_attrs);

   if (is_4x4_board())
      total_num_attrs += ARRAY_SIZE(reg_4x4_attrs);

   if (is_fzp_db_board() || is_fzm_dbo_1g_board() || is_fzm_dbo_board())
      total_num_attrs += ARRAY_SIZE(reg_rfcard_attrs);

   /* Add 1 more for NULL */
   total_num_attrs++;

   /* No need to free upon driver exit with devm_kzalloc */
   cpld_attrs = devm_kzalloc(dev, total_num_attrs * sizeof(struct attribute *), GFP_KERNEL);
   if (!cpld_attrs)
      goto out_device;

   cpld_attrs_p = cpld_attrs;

   /* Populate the allocated cpld_attrs with pointers to all of the
    * attributes we want to add */

   for (i = 0; i < ARRAY_SIZE(reg_base_attrs); i++) {
      *cpld_attrs_p = &reg_base_attrs[i].attr.attr,
       cpld_attrs_p++;
   }

   for (i = 0; i < ARRAY_SIZE(led_attrs); i++) {
      *cpld_attrs_p = &led_attrs[i].attr,
       cpld_attrs_p++;
   }

   if (is_wifi_supported()) {
      for (i = 0; i < ARRAY_SIZE(reg_wifi_attrs); i++) {
         *cpld_attrs_p = &reg_wifi_attrs[i].attr.attr,
          cpld_attrs_p++;
      }
   }

   if (is_fzp_board()) {
      for (i = 0; i < ARRAY_SIZE(reg_fzp_attrs); i++) {
         *cpld_attrs_p = &reg_fzp_attrs[i].attr.attr,
          cpld_attrs_p++;
      }
   }

   if (is_fzp_board() && is_wifi_supported()) {
      for (i = 0; i < ARRAY_SIZE(reg_fzp_wifi_attrs); i++) {
         *cpld_attrs_p = &reg_fzp_wifi_attrs[i].attr.attr,
          cpld_attrs_p++;
      }
   }

   if (is_4x4_board()) {
      for (i = 0; i < ARRAY_SIZE(reg_4x4_attrs); i++) {
         *cpld_attrs_p = &reg_4x4_attrs[i].attr.attr,
          cpld_attrs_p++;
      }
   }

   if (is_fzp_db_board() || is_fzm_dbo_1g_board() || is_fzm_dbo_board()) {
      for (i = 0; i < ARRAY_SIZE(reg_rfcard_attrs); i++) {
         *cpld_attrs_p = &reg_rfcard_attrs[i].attr.attr,
          cpld_attrs_p++;
      }
   }

   /* Terminate list of attribute */
   *cpld_attrs_p = NULL;

   cpld_attr_group.attrs = cpld_attrs;

   err = sysfs_create_group(&dev->kobj, &cpld_attr_group);
   if (err)
      goto out_device;

   err = cpld_irqnotify_probe(dev);
   if (err)
      goto out_sysfs_group;

   /* Set the FPGA2 bit in GOR/GDR */
   cpld_set_gpio();

   cpldusr_proc = proc_create(DRV_NAME, 0444, NULL, &proc_ops);
   if (!cpldusr_proc) {
      MOD_ERROR("failed to create /proc/%s\n", DRV_NAME);
      goto out_cpld_irqnotify;
   }

   err = cpld_sfp_probe(dev);
   if (err)
      goto out_proc;

   return 0;

out_proc:
   remove_proc_entry(DRV_NAME, NULL);
out_cpld_irqnotify:
   cpld_irqnotify_remove(dev);
out_sysfs_group:
   sysfs_remove_group(&dev->kobj, &cpld_attr_group);
out_device:
   device_unregister(dev);
out_class:
   class_unregister(&cpld_class);
out_cdev:
   cdev_del(&cpldusr_cdev);
out_chrdev_region:
   unregister_chrdev_region(devt, 1);

   return err;
}

/**
 * @internal
 * @brief Called when the module is removed from the kernel.
 */
int __exit cpld_usr_remove(struct device *pdev)
{
   dev_t devt = dev->devt;

   /* printk("<2> Bye, cruel world\n"); */
#ifdef CPLD_DEBUG
   printk(KERN_EMERG "in cpldusr_remove\n");
#endif

   remove_proc_entry(DRV_NAME, NULL);

   cpld_irqnotify_remove(dev);
   sysfs_remove_group(&dev->kobj, &cpld_attr_group);
   device_unregister(dev);
   class_unregister(&cpld_class);
   cdev_del(&cpldusr_cdev);
   unregister_chrdev_region(devt, 1);

   return 0;
}

//HACK
//This function should be removed
//It is currently required to ensure that the CPLD interrupt controller
//exists before anything tries to use it.
//This is resolved in 3.15+ with EPROBE_DEFER on IRQs
extern void make_dependency_to_cpld_usr(void);
void make_dependency_to_cpld_usr(void)
{
#ifdef CPLD_DEBUG
   printk(KERN_EMERG "CPLD is loaded!\n");
#endif
}
EXPORT_SYMBOL(make_dependency_to_cpld_usr);
