/* Copyright (c) 2014 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    SRID       Description
 * 04/12/2013 yongxu               Initial creation from lsp_k2_driver
 * </pre>
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/spi/spi.h>
#include <linux/hardirq.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/mutex.h>
#include <linux/interrupt.h>
#include <linux/of_irq.h>
#include <linux/irqdomain.h>
#include <asm/uaccess.h>
#include <linux/platform_device.h>
#include "fzm_fpga.h"
#include "fzm_fpga_eis.h"
#include "fpga_internal.h"


static struct mutex *fpga_get_lock(struct device *dev)
{
   struct spi_device *spi = to_spi_device(dev);
   struct fpga_info *info = spi_get_drvdata(spi);

   return &info->mutex;
}

void fpga_lock(struct device *dev)
{
   struct mutex *mutex = fpga_get_lock(dev);

   mutex_lock(mutex);
}
EXPORT_SYMBOL(fpga_lock);


int fpga_lock_interruptible(struct device *dev)
{
   struct mutex *mutex = fpga_get_lock(dev);

   return mutex_lock_interruptible(mutex);
}
EXPORT_SYMBOL(fpga_lock_interruptible);


int fpga_trylock(struct device *dev)
{
   struct mutex *mutex = fpga_get_lock(dev);

   return mutex_trylock(mutex);
}
EXPORT_SYMBOL(fpga_trylock);

void fpga_unlock(struct device *dev)
{
   struct mutex *mutex = fpga_get_lock(dev);

   mutex_unlock(mutex);
}
EXPORT_SYMBOL(fpga_unlock);

struct fpga_poll_data {
   struct spi_message m;
   u8 *buf;
   unsigned long timeout;
   struct completion done;
};

/* SPI xfer complete callback which polls for a non-zero response */
static void fpga_poll_cb(void *arg)
{
   struct fpga_poll_data *d = arg;
   int err;

   if (d->m.status) {
      /* error */
      goto finish;
   }

   if (d->buf[0]) {
      /* polling is complete */
      goto finish;
   }

   /* for the first retry, remove the first write and just keep the poll */
   if (!list_is_last(d->m.transfers.next, &d->m.transfers)) {
      list_del(d->m.transfers.next);
   } else {
      /* in case of preemption, check timeout only if we haven't polled a second time */
      if (time_is_before_eq_jiffies(d->timeout)) {
         d->m.status = -ETIMEDOUT;
         goto finish;
      }
   }

   /* kick off the poll again */
   err = spi_async(d->m.spi, &d->m);
   if (err) {
      d->m.status = err;
      goto finish;
   }
   return;

finish:
   complete(&d->done);
}

/**
 * fpga_rw
 * @dev:     target device (FPGA)
 * @devid:   0 for FPGA, 1 for CAT, 2 for PLL
 * @mt:      message type
 * @pld_0:   payload0
 * @pld_1:   payload1
 * @data:    pointer to the buf where the read data to be put
 *
 * Cannot be called from interrupt context.
 *
 * Returns 0 upon success, n
 */
int fpga_rw(struct device *dev,
            FPGA_SPI_DEVID_t devid,
            FPGA_SPI_MT_t mt,
            u32 pld_0,
            u32 pld_1,
            u32 *data)
{
   struct fpga_poll_data d;
   struct spi_transfer t[2];
   struct spi_device *spi = to_spi_device(dev);
   int ret;
   u8 buffer[12]; /* 9 bytes for data, 3 for alignment */
   u8 *buf = &buffer[3];
   bool is_read = false;

   /* lock must be held to prevent burst operations from being interrupted */
   WARN_ON(!mutex_is_locked(fpga_get_lock(dev)));

   /* spi_sync below cannot be called from interrupt context */
   BUG_ON(in_interrupt());

#ifdef FPGA_TEST
   dev_info(dev, "did: %d, mt: %d, p0: %x, p1: %x\n", devid, mt, pld_0, pld_1);
#endif

   // stub out the fpga_rw function
   FUNC_TRACE;
   /* Command Construction*/
   buf[0]  = (mt & FPGA_SPI_MT_MASK) << FPGA_SPI_MT_SHIFT;
   buf[0] |= (devid & FPGA_SPI_DEVID_MASK) << FPGA_SPI_DEVID_SHIFT;

   *((__be32 *)&buf[1]) = cpu_to_be32(pld_0);
   *((__be32 *)&buf[5]) = cpu_to_be32(pld_1);

   if ((FPGA_SPI_MT_GEN_READ_BRST == mt) ||
         (FPGA_SPI_MT_GEN_READ == mt) ||
         (FPGA_SPI_MT_DEV_READ == mt)) {
      is_read = true;
   }

   memset(t, 0x00, sizeof(t));
   spi_message_init_with_transfers(&d.m, t, 2);
   init_completion(&d.done);
   d.timeout = jiffies + HZ / 2; /* 500ms */
   d.buf = buf;

   /* setup our custom callback */
   d.m.complete = fpga_poll_cb;
   d.m.context = &d;

   /* command/data write */
   t[0].tx_buf = buf;
   t[0].len = 9;
   t[0].cs_change = 1;

   /* polling xfer: tx will clock out 0x00 */
   /* in most cases, this will succeed the first time */
   t[1].rx_buf = buf;
   t[1].len = is_read ? 9 : 1;

again:
   /* cannot use spi_bus_lock safely here since it sets exclusive access and
    * will cause our unlocked spi_async to fail */
   mutex_lock(&spi->master->bus_lock_mutex);
   ret = spi_async_locked(spi, &d.m);
   mutex_unlock(&spi->master->bus_lock_mutex);

   if (ret) {
      dev_err(dev, "spi_async_locked failed: %d\n", ret);
      return ret;
   }

   wait_for_completion(&d.done);
   ret = d.m.status;

   if (ret) {
      dev_err(dev, "spi xfer failed: %d\n", ret);
      if (ret == -EBUSY) {
         /* corner case where someone locked the bus for exclusive access (spi_bus_lock) */
         goto again;
      }
      return ret;
   }

   if (is_read) {
      /* get_unaligned() should be used to be generic, but that isn't
       * optimized until a later kernel. ARMv7 can read/write unaligned just
       * fine and has a byteswap instruction, but these should always end up
       * aligned anyway.
       */
      if (FPGA_SPI_MT_GEN_READ_BRST == mt) {
         *data++ = be32_to_cpup((__be32 *)&buf[1]);
      }
      *data = be32_to_cpup((__be32 *)&buf[5]);

#ifdef FPGA_TEST
      dev_info(dev, "returned value: %x\n", *((u32 *)data));
#endif
   }

   return 0;
}
EXPORT_SYMBOL(fpga_rw);

int fpga_read(struct device *dev, u32 addr, u32 *val, FPGA_SPI_DEVID_t dev_id)
{
   int ret;

   fpga_lock(dev);
   ret = fpga_read_unlocked(dev, addr, val, dev_id, false);
   fpga_unlock(dev);

   return ret;
}
EXPORT_SYMBOL(fpga_read);

int fpga_write(struct device *dev, u32 addr, u32 val, FPGA_SPI_DEVID_t dev_id)
{
   int ret;

   fpga_lock(dev);
   ret = fpga_write_unlocked(dev, addr, &val, dev_id, false);
   fpga_unlock(dev);

   return ret;
}
EXPORT_SYMBOL(fpga_write);

int fpga_read_unlocked(struct device *dev, u32 addr, u32 *val, FPGA_SPI_DEVID_t dev_id, bool is_burst)
{
   FPGA_SPI_MT_t mt;

   if (!is_burst) {
      mt = FPGA_SPI_MT_GEN_READ;
   } else {
      mt = FPGA_SPI_MT_GEN_READ_BRST;
   }

   FUNC_TRACE;

   return fpga_rw(dev, dev_id, mt, addr, 0, val);
}
EXPORT_SYMBOL(fpga_read_unlocked);

int fpga_read_block_unlocked(struct device *dev, u32 addr, u32 *val, FPGA_SPI_DEVID_t dev_id, unsigned count)
{
   int err;

   if (unlikely(!count)) {
      return 0;
   }

   /* do single reads until we end up with a multiple of 8 */
   do {
      err = fpga_read_unlocked(dev, addr, val, dev_id, false);
      if (err)
         goto out;
      addr += sizeof(*val);
      count--;
      val++;
   } while (count % 2);

   /* handle rest with bursts */
   if (unlikely(count)) {
      while (count) {
         err = fpga_read_unlocked(dev, 0, val, dev_id, true);
         if (err)
            goto out;
         count -= 2;
         val += 2;
      }
   }

out:
   return err;
}
EXPORT_SYMBOL(fpga_read_block_unlocked);

int fpga_write_unlocked(struct device *dev, u32 addr, const u32 *val, FPGA_SPI_DEVID_t dev_id, bool is_burst)
{
   FPGA_SPI_MT_t mt;
   u32 p0, p1;

   if (!is_burst) {
      mt = FPGA_SPI_MT_GEN_WRITE;
      p0 = addr;
      p1 = *val;
   } else {
      mt = FPGA_SPI_MT_GEN_WRITE_BRST;
      p0 = *val;
      p1 = *(val + 1);
   }

   FUNC_TRACE;

   return fpga_rw(dev, dev_id, mt, p0, p1, NULL);
}
EXPORT_SYMBOL(fpga_write_unlocked);

int fpga_write_block_unlocked(struct device *dev, u32 addr, const u32 *val, FPGA_SPI_DEVID_t dev_id, unsigned count)
{
   int err;

   if (unlikely(!count)) {
      return 0;
   }

   /* do single writes until we end up with a multiple of 8 */
   do {
      err = fpga_write_unlocked(dev, addr, val, dev_id, false);
      if (err)
         goto out;
      addr += sizeof(*val);
      count--;
      val++;
   } while (count % 2);

   /* handle rest with bursts */
   if (unlikely(count)) {
      while (count) {
         err = fpga_write_unlocked(dev, 0, val, dev_id, true);
         if (err)
            goto out;
         count -= 2;
         val += 2;
      }
   }

out:
   return err;
}
EXPORT_SYMBOL(fpga_write_block_unlocked);

int fpga_modify_unlocked(struct device *dev, u32 addr, u32 mask, u32 val, FPGA_SPI_DEVID_t dev_id)
{
   u32 cur_val;
   int err;
   err = fpga_read_unlocked(dev, addr, &cur_val, FPGA_SPI_DEVID_N_A + dev_id, false);
   if (err == 0) {
      cur_val = (cur_val & ~mask) | (val & mask);
      err = fpga_write_unlocked(dev, addr, &cur_val, FPGA_SPI_DEVID_N_A + dev_id, false);
   }
   return err;
}
EXPORT_SYMBOL(fpga_modify_unlocked);

int fpga_modify(struct device *dev, u32 addr, u32 mask, u32 val, FPGA_SPI_DEVID_t dev_id)
{
   int err;
   fpga_lock(dev);
   err = fpga_modify_unlocked(dev, addr, mask, val, dev_id);
   fpga_unlock(dev);
   return err;
}
EXPORT_SYMBOL(fpga_modify);


static int fzm_fpga_probe(struct spi_device *spi)
{
   int err;
   struct fpga_info *info;

   info = kmalloc(sizeof(*info), GFP_KERNEL);
   if (!info) {
      dev_err(&spi->dev, "kmalloc failed\n");
      return -ENOMEM;
   }

   mutex_init(&info->mutex);

   spi_set_drvdata(spi, info);

   err = fpga_bus_probe(&spi->dev);
   if (err) {
      goto out_info;
   }

   err = fpga_usr_probe(&spi->dev);
   if (err) {
      goto out_bus;
   }

   err = fpga_ic_probe(&spi->dev);
   if (err)
      goto out_usr;

   fpga_bus_register_devices(&spi->dev);

   return 0;

out_usr:
   fpga_usr_remove(&spi->dev);
out_bus:
   fpga_bus_remove(&spi->dev);
out_info:
   kfree(info);

   return err;
}

static int fzm_fpga_remove(struct spi_device *spi)
{
   int err;

   fpga_bus_unregister_devices(&spi->dev);

   err = fpga_ic_remove(&spi->dev);

   err = fpga_usr_remove(&spi->dev);

   err = fpga_bus_remove(&spi->dev);

   kfree(spi_get_drvdata(spi));

   return err;
}

static const struct of_device_id fzm_fpga_id[] = {
   { .compatible = "nsn,fzm-fpga_usr" },
   { }
};
MODULE_DEVICE_TABLE(of, fzm_fpga_id);

MODULE_ALIAS("spi:fzm-fpga_usr");

static struct spi_driver fzm_fpga_driver = {
   .driver = {
      .name = "fzm-fpga_usr",
      .owner = THIS_MODULE,
      .of_match_table = of_match_ptr(fzm_fpga_id),
   },
   .probe = fzm_fpga_probe,
   .remove = fzm_fpga_remove,
};

module_spi_driver(fzm_fpga_driver);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nokia Solutions and Networks");
MODULE_DESCRIPTION("FZM FPGA_USR SPI driver for NSN FZM");
